use dotenvy::dotenv;
use ldap3::{LdapConn, Mod};
use skynet_ldap_backend::{db_init, get_config, read_csv, Accounts, Config, get_now_iso};
use sqlx::{Pool, Sqlite};
use std::{collections::HashSet, env, error::Error};

#[async_std::main]
async fn main() -> tide::Result<()> {
    let config = get_config();

    update_users(&config).await?;
    update_admin(&config).await?;
    update_committee(&config).await?;

    Ok(())
}

async fn update_users(config: &Config) -> tide::Result<()> {
    let mut users_tmp = HashSet::new();
    // default user to ensure group is never empty
    users_tmp.insert(String::from("compsoc"));

    if let Ok(x) = env::var("USERS_LIFETIME") {
        for user in x.split(',').collect::<Vec<&str>>() {
            users_tmp.insert(user.to_string());
        }
    }

    /*
    pull in data from wolves (csv or api (hopefully api)
    pull entire ldap data

    for every valid user in wolves match to ldap
    add to users
    */
    // pull from wolves csv
    for user in from_csv(config).await.unwrap_or_default() {
        users_tmp.insert(user);
    }

    // sorting makes it easier/faster
    if let Ok(x) = env::var("USERS_BANNED") {
        for user in x.split(',').collect::<Vec<&str>>() {
            users_tmp.remove(user);
        }
    }

    // easier to work with Strings above but easier to work with &str below
    let users: Vec<&str> = users_tmp.iter().map(|s| &**s).collect();

    update_group(config, "skynet-users", &users, true).await?;

    Ok(())
}

fn uid_to_dn(uid: &str) -> String {
    format!("uid={},ou=users,dc=skynet,dc=ie", uid)
}

async fn update_admin(config: &Config) -> tide::Result<()> {
    dotenv().ok();

    // read from teh env
    if let Ok(x) = env::var("USERS_ADMIN") {
        let users = x.split(',').collect::<Vec<&str>>();

        update_group(config, "skynet-admins", &users, true).await?;
        // admins automatically get added as users
        update_group(config, "skynet-users", &users, false).await?;
    }
    Ok(())
}

async fn update_committee(config: &Config) -> tide::Result<()> {
    dotenv().ok();

    // read from teh env
    if let Ok(x) = env::var("USERS_COMMITTEE") {
        let users = x.split(',').collect::<Vec<&str>>();

        update_group(config, "skynet-committee", &users, true).await?;
        // admins automatically get added as users
        update_group(config, "skynet-users", &users, false).await?;
    }
    Ok(())
}

async fn update_group(config: &Config, group: &str, users: &[&str], replace: bool) -> tide::Result<()> {
    if users.is_empty() {
        return Ok(());
    }

    let mut ldap = LdapConn::new(&config.ldap_host)?;

    // use the admin account
    ldap.simple_bind(&config.ldap_admin, &config.ldap_admin_pw)?.success()?;

    let dn = format!("cn={},ou=groups,dc=skynet,dc=ie", group);
    let members = users.iter().map(|uid| uid_to_dn(uid)).collect();
    let mods = if replace {
        vec![Mod::Replace("member".to_string(), members)]
    } else {
        vec![Mod::Add("member".to_string(), members)]
    };

    if let Err(x) = ldap.modify(&dn, mods) {
        println!("{:?}", x);
    }

    let dn_linux = format!("cn={}-linux,ou=groups,dc=skynet,dc=ie", group);
    let members_linux = users.iter().map(|uid| uid.to_string()).collect();
    let mods = if replace {
        vec![Mod::Replace("memberUid".to_string(), members_linux)]
    } else {
        vec![Mod::Add("memberUid".to_string(), members_linux)]
    };
    if let Err(x) = ldap.modify(&dn_linux, mods) {
        println!("{:?}", x);
    };

    // tidy up
    ldap.unbind()?;

    Ok(())
}

async fn from_csv(config: &Config) -> Result<HashSet<String>, Box<dyn Error>> {
    let db = db_init(config).await.unwrap();

    let mut uids = HashSet::new();

    let records = read_csv(config)?;

    for record in records {
        // only import users if it is actually active.
        if record.expiry < get_now_iso(true) {
            continue;
        }

        if let Some(uid) = account_mail_get_uid(&db, &record.email).await {
            uids.insert(uid);
        } else if let Some(uid) = account_id_get_uid(&db, &record.id_student).await {
            uids.insert(uid);
        }
    }

    Ok(uids)
}

async fn account_mail_get_uid(db: &Pool<Sqlite>, mail: &str) -> Option<String> {
    match sqlx::query_as::<_, Accounts>(
        r#"
        SELECT user
        FROM accounts
        WHERE mail == ?
        "#,
    )
    .bind(mail)
    .fetch_all(db)
    .await
    {
        Ok(res) => {
            if res.is_empty() {
                None
            } else {
                Some(res[0].user.to_owned())
            }
        }
        Err(_) => None,
    }
}

async fn account_id_get_uid(db: &Pool<Sqlite>, id: &str) -> Option<String> {
    match sqlx::query_as::<_, Accounts>(
        r#"
        SELECT user
        FROM accounts
        WHERE student_id == ?
        "#,
    )
    .bind(id)
    .fetch_all(db)
    .await
    {
        Ok(res) => {
            if res.is_empty() {
                None
            } else {
                Some(res[0].student_id.to_owned())
            }
        }
        Err(_) => None,
    }
}
